from typing import Union, List, Any

from treepath.path.traverser.traverser_match import TraverserMatch
from treepath.path.vertex.vertex import Vertex


class Match:
    """

    """
    __slots__ = '_traverser_match'

    def __init__(self,
                 traverser_match: TraverserMatch,
                 ):
        self._traverser_match = traverser_match

    @property
    def path_as_list(self) -> List[Any]:
        """

        """
        return [Match(traverser_match) for traverser_match in self._traverser_match.path_as_list]

    @property
    def path(self) -> str:
        """

        """
        return self._traverser_match.path

    @property
    def path_segment(self):
        """

        """
        return self._traverser_match.path_segment

    @property
    def parent(self) -> Union[Any, None]:
        """

        """
        parent = self._traverser_match.parent
        if parent:
            return Match(parent)
        else:
            return None

    @property
    def data_name(self) -> Union[str, int]:
        """

        """
        return self._traverser_match.data_name

    @property
    def data(self) -> Union[dict, list, str, int, float, bool, None]:
        """

        """
        return self._traverser_match.data

    @property
    def vertex(self) -> Vertex:
        """

        """
        return self._traverser_match.vertex

    def __repr__(self):
        return repr(self._traverser_match)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return self is other \
               or isinstance(other, Match) and \
               ((self._traverser_match is other._traverser_match)
                or (self.data == other.data) and (self.parent == other.parent)
                )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.path)



from treepath.path.vertex.vertex import Vertex


class TreepathException(Exception):
    """Basic exception for errors raised by treepath"""

    def __init__(self, vertex: Vertex):
        self.vertex = vertex
        self.is_msg_resolved = False
        super().__init__()

    def __repr__(self):
        if self.is_msg_resolved:
            return self.msg
        self.msg = self._resolve_msg()
        return self.msg

    def __str__(self):
        return self.__repr__()

    def _resolve_msg(self):
        self.is_msg_resolved = True
        path = repr(self.vertex)
        return f"TreepathException(path: {path})"



"""
# The **treepath** Package

**treepath** uses [declarative programming](https://en.wikipedia.org/wiki/Declarative_programming) approach for
extracting data from a [json](https://docs.python.org/3/library/json.html) data structure.  The expressions are a
[query language](https://en.wikipedia.org/wiki/Query_language) similar to
[jsonpath](https://goessner.net/articles/JsonPath/), and [Xpath](https://en.wikipedia.org/wiki/XPath), but are
written in native python syntax.

# Quick start

```python
    from treepath import path, get
    data = {
        "a": {
            "b": {
                "c": 1
            }
        }
    }
    value = get(path.a.b.c, data)
    assert value == 1
```
"""
from treepath.path.builder.dash_path_builder import DashPathRoot
from treepath.path.builder.patch_constants import rec
from treepath.path.builder.patch_constants import recursive
from treepath.path.builder.patch_constants import wc
from treepath.path.builder.patch_constants import wildcard
from treepath.path.builder.root_path_builder import RootPathBuilder
from treepath.path.exceptions.infinite_loop_detected import InfiniteLoopDetected
from treepath.path.exceptions.match_not_found_error import MatchNotFoundError
from treepath.path.exceptions.nested_match_not_found_error import NestedMatchNotFoundError
from treepath.path.exceptions.path_syntax_error import PathSyntaxError
from treepath.path.exceptions.stop_traversing import StopTraversing
from treepath.path.exceptions.traversing_error import TraversingError
from treepath.path.exceptions.treepath_exception import TreepathException
from treepath.path.traverser.match import Match
from treepath.path.traverser.trace import Trace
from treepath.path.traverser.trace import log_to
from treepath.path.traverser.traverser_functions import find
from treepath.path.traverser.traverser_functions import find_matches
from treepath.path.traverser.traverser_functions import get
from treepath.path.traverser.traverser_functions import get_match
from treepath.path.traverser.traverser_functions import has
from treepath.path.traverser.traverser_functions import has_all
from treepath.path.traverser.traverser_functions import has_any
from treepath.path.traverser.traverser_functions import has_not
from treepath.path.traverser.traverser_functions import nested_find_matches
from treepath.path.traverser.traverser_functions import nested_get_match

path = RootPathBuilder()
pathd = DashPathRoot()



from abc import abstractmethod, ABC


class AbstractPathBuilder(ABC):
    __slots__ = ()

    @abstractmethod
    def create_path_builder(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def tranform_attribute_name(self, name):
        raise NotImplementedError



from treepath.path.traverser.traverser_match import TraverserMatch


class ListMatch(TraverserMatch):
    __slots__ = ()

    @property
    def path_segment(self):
        return f"[{self.real_data_name}]"



from typing import Union, Iterator, Callable, Any, Tuple

from treepath.path.builder.path_builder import PathBuilder, get_vertex_from_path_builder
from treepath.path.builder.path_builder_predicate import PathBuilderPredicate
from treepath.path.builder.path_predicate import PathPredicate
from treepath.path.exceptions.match_not_found_error import MatchNotFoundError
from treepath.path.exceptions.nested_match_not_found_error import NestedMatchNotFoundError
from treepath.path.traverser.has_function import create_has_predicate
from treepath.path.traverser.match import Match
from treepath.path.traverser.match_traverser import MatchTraverser
from treepath.path.traverser.nested_match_traverser import NestedMatchTraverser
from treepath.path.traverser.nested_value_traverser import NestedValueTraverser
from treepath.path.traverser.predicate_match import PredicateMatch
from treepath.path.traverser.trace import Trace
from treepath.path.traverser.value_traverser import ValueTraverser
from treepath.path.util.decorator import pretty_repr, add_attr

_has_typing_first_arg = Union[
    PathBuilderPredicate,
    PathPredicate,
    Callable[[Match], Any]
]
_has_typing_single_arg_functions = Callable[[Any], Any]

_has_tuple_arg_type = Tuple[
    _has_typing_first_arg,
    _has_typing_single_arg_functions,
]

_has_multiple_arg_type = Union[
    _has_typing_first_arg,
    _has_tuple_arg_type,
]

_not_set = dict()


def get(
        expression: PathBuilder,
        data: Union[dict, list, Match],
        default=_not_set,
        trace: Callable[[Trace], None] = None
) -> Union[dict, list, str, int, float, bool, None]:
    """

    """
    must_match = (default is _not_set)
    match = get_match(expression, data, must_match=must_match, trace=trace)
    if match:
        return match.data
    return default


def find(
        expression: PathBuilder,
        data: Union[dict, list, Match],
        trace: Callable[[Trace], None] = None
) -> Iterator[Union[dict, list, str, int, float, bool, None]]:
    """

    """
    if isinstance(data, Match):
        return nested_find(expression, data, trace=trace)

    vertex = get_vertex_from_path_builder(expression)
    traverser = ValueTraverser(data, vertex, trace=trace)
    traverser_iter = iter(traverser)
    return traverser_iter


def get_match(
        expression: PathBuilder,
        data: Union[dict, list, Match],
        must_match: bool = True,
        trace: Callable[[Trace], None] = None
) -> Union[Match, None]:
    """

    """
    if isinstance(data, Match):
        return nested_get_match(expression, data, must_match=must_match, trace=trace)

    vertex = get_vertex_from_path_builder(expression)
    traverser = MatchTraverser(data, vertex, trace=trace)
    traverser_iter = iter(traverser)
    try:
        return next(traverser_iter)
    except StopIteration:
        pass
    if must_match:
        raise MatchNotFoundError(vertex)
    return None


def find_matches(
        expression: PathBuilder,
        data: Union[dict, list, Match],
        trace: Callable[[Trace], None] = None
) -> Iterator[Match]:
    """

    """
    if isinstance(data, Match):
        return nested_find_matches(expression, data, trace=trace)

    vertex = get_vertex_from_path_builder(expression)
    traverser = MatchTraverser(data, vertex, trace=trace)
    traverser_iter = iter(traverser)
    return traverser_iter


def nested_find(
        expression: PathBuilder,
        parent_match: Match,
        trace: Callable[[Trace], None] = None
) -> Iterator[Union[dict, list, str, int, float, bool, None]]:
    """

    """
    vertex = get_vertex_from_path_builder(expression)
    traverser = NestedValueTraverser(parent_match._traverser_match, vertex, trace=trace)
    traverser_iter = iter(traverser)
    return traverser_iter


def nested_get_match(
        expression: PathBuilder,
        parent_match: Match,
        must_match: bool = True,
        trace: Callable[[Trace], None] = None
) -> Union[Match, None]:
    """

    """
    if isinstance(parent_match, PredicateMatch):
        trace = parent_match.trace

    vertex = get_vertex_from_path_builder(expression)
    traverser = NestedMatchTraverser(parent_match._traverser_match, vertex, trace=trace)
    traverser_iter = iter(traverser)
    try:
        return next(traverser_iter)
    except StopIteration:
        pass
    if must_match:
        raise NestedMatchNotFoundError(parent_match, vertex)
    return None


def nested_find_matches(
        expression: PathBuilder,
        parent_match: Match,
        trace: Callable[[Trace], None] = None
) -> Iterator[Match]:
    """

    """
    if isinstance(parent_match, PredicateMatch):
        trace = parent_match.trace
    vertex = get_vertex_from_path_builder(expression)
    traverser = NestedMatchTraverser(parent_match._traverser_match, vertex, trace=trace)
    traverser_iter = iter(traverser)
    return traverser_iter


def has_these(*args: _has_multiple_arg_type, repr_join_key=', '):
    """

    """

    def process_has_arg(arg):
        if isinstance(arg, tuple):
            return create_has_predicate(nested_find_matches, *arg)
        else:
            return create_has_predicate(nested_find_matches, arg)

    has_predicates = [process_has_arg(arg) for arg in args]

    def wrap(function):
        @pretty_repr(lambda: f"{repr_join_key.join(map(repr, has_predicates))}")
        def predicate(parent_match):
            return function(parent_match, *has_predicates)

        return predicate

    return wrap


def has_all(*args: _has_multiple_arg_type):
    """
    Tuple[Union[PathBuilderPredicate, PathPredicate, Callable[[Match], Any]]]
    """

    @has.these(*args, repr_join_key=' and ')
    def and_predicate(parent_match: Match, *predicates) -> Any:
        for predicate in predicates:
            if not predicate(parent_match):
                return False
        return True

    return and_predicate


def has_any(*args: _has_multiple_arg_type):
    """

    """

    @has.these(*args, repr_join_key=' or ')
    def or_predicate(parent_match: Match, *predicates) -> Any:
        for predicate in predicates:
            if predicate(parent_match):
                return True
        return False

    return or_predicate


def has_not(
        path: _has_typing_first_arg,
        *single_arg_functions: _has_typing_single_arg_functions) -> Callable[[Match], Any]:
    """
    Tuple[Union[PathBuilderPredicate, PathPredicate, Callable[[Match], Any]]]
    """

    predicate = create_has_predicate(nested_find_matches, path, *single_arg_functions)

    def not_predicate(parent_match: Match) -> Any:
        return not predicate(parent_match)

    return not_predicate


@add_attr("these", has_these)
def has(
        path: _has_typing_first_arg,
        *single_arg_functions: _has_typing_single_arg_functions) -> Callable[[Match], Any]:
    """

    """
    return create_has_predicate(nested_find_matches, path, *single_arg_functions)



from treepath.path.traverser.traverser_match import TraverserMatch


class KeyMatch(TraverserMatch):
    __slots__ = ()

    @property
    def path_segment(self):
        return f".{self.real_data_name}"



from typing import Callable

from treepath.path.exceptions.infinite_loop_detected import InfiniteLoopDetected
from treepath.path.exceptions.stop_traversing import StopTraversing
from treepath.path.traverser.match import Match
from treepath.path.traverser.root_match import RootMatch
from treepath.path.traverser.trace import Trace
from treepath.path.traverser.traverser_match import TraverserMatch
from treepath.path.vertex.vertex import Vertex


class MatchTraverser:
    __slots__ = 'vertex_path', \
                'leaf_vertex', \
                'root_data', \
                '_invoke_next_action', \
                'current_match', \
                'root_match', \
                'trace'

    def __init__(self,
                 root_data,
                 leaf_vertex: Vertex,
                 trace: Callable[[Trace], None] = None
                 ):
        self.vertex_path = leaf_vertex.path_as_list
        self.leaf_vertex = leaf_vertex
        self.root_data = root_data
        self._invoke_next_action = self.done_action
        self.current_match = None
        self.root_match = None
        self.trace = trace

    def __iter__(self):
        self._invoke_next_action = self.init_action
        return self

    def __next__(self) -> Match:
        result = None
        count = 1000000
        while result is None:
            result = self._invoke_next_action()
            count += -1
            if not count:
                raise InfiniteLoopDetected(
                    self.current_match,
                    "A 1000000000 iteration have occurred since the __next__ last return.   Possible infinite loop "
                )
        return Match(self.current_match)

    def remember_on_catch(self, match: TraverserMatch, remembered_catch_state):
        match.remembered_catch_state = remembered_catch_state
        match.remembered_on_catch_match = match
        match.remembered_on_catch_action = self.match_action

    def restore_on_catch(self, match: TraverserMatch):
        match.remembered_catch_state = None
        if self.root_match == match:
            match.remembered_on_catch_match = None
            match.remembered_on_catch_action = self.done_action()
        else:
            parent = match.real_parent
            match.remembered_on_catch_match = parent.remembered_on_catch_match
            match.remembered_on_catch_action = parent.remembered_on_catch_action

    def init_action(self):
        vertex = self.vertex_path[0]
        root_match = RootMatch(
            None,  # real_parent
            vertex.name,
            self.root_data,  # data
            vertex,  # vertex
            0,  # vertex_index
            None,  # remembered_on_catch_match
            self.done_action
        )
        root_match.remembered_on_catch_match = root_match
        self.root_match = root_match
        self.current_match = root_match
        self._invoke_next_action = self.report_action

    def match_action(self):
        current_match = self.current_match
        next_vertex_index = current_match.vertex_index
        next_vertex_index += 1
        next_vertex = self.vertex_path[next_vertex_index]

        # apply the get_match
        next_match = next_vertex.match(current_match, self, next_vertex_index)

        if self.trace:
            self.trace(
                Trace(
                    Match( current_match),
                    Match(next_match) if next_match else None,
                    next_vertex
                )
            )

        if next_match:
            self.current_match = next_match
            self._invoke_next_action = self.report_action
        else:
            self.current_match = current_match.remembered_on_catch_match
            self._invoke_next_action = current_match.remembered_on_catch_action

    def report_action(self):
        current_match = self.current_match
        if current_match.vertex == self.leaf_vertex:
            self._invoke_next_action = self.catch_action
            return True
        else:
            self._invoke_next_action = self.match_action

    def catch_action(self):
        current_match = self.current_match
        self.current_match = current_match.remembered_on_catch_match
        self._invoke_next_action = current_match.remembered_on_catch_action

    def done_action(self):
        raise StopTraversing(self.leaf_vertex)



from treepath.path.exceptions.treepath_exception import TreepathException


class MatchNotFoundError(TreepathException, LookupError):
    """
    MatchNotFoundError is raised when a get_match is expected; however, the data structure has no more element to traverse
    """

    def _resolve_msg(self):
        path = repr(self.vertex)
        return f"MatchNotFoundError(No get_match occurred on path: {path})"



from typing import Callable

from treepath.path.traverser.match import Match
from treepath.path.vertex.vertex import Vertex


class Trace:
    __slots__ = '_last_match', '_next_match', '_next_vertex', '_predicate_match'

    def __init__(self, last_match: Match, next_match: Match, next_vertex: Vertex):
        self._last_match = last_match
        self._next_match = next_match
        self._next_vertex = next_vertex
        self._predicate_match = None

    @property
    def last_match(self) -> Match:
        return self._last_match

    @property
    def next_match(self) -> Match:
        return self._next_match

    @property
    def next_vertex(self) -> Vertex:
        return self._next_vertex

    @property
    def predicate_match(self) -> Match:
        return self._predicate_match

    @predicate_match.setter
    def predicate_match(self, predicate_match: Match):
        if not self._predicate_match:
            self._predicate_match = predicate_match


def _log(log: Callable[[str], None], trace: Trace):
    last_path = trace.last_match.path
    vertex_path_segment = trace.next_vertex.path_segment
    if trace.next_match:
        next_match_path_segment = trace.next_match.path_segment
        data = repr(trace.next_match.data)
        trunc_data = data if len(data) < 20 else data[:20] + '...'
        result = f"{trunc_data}"
    else:
        result = 'no match'
    message = f" at {last_path}{vertex_path_segment} got {result}"

    if trace.predicate_match:
        predicate_path = f" at {trace.predicate_match.path}"
        message = message.replace(predicate_path, " has ".rjust(len(predicate_path)), 1)

    log(message)


def log_to(out: Callable[[str], None]):
    def to_out(trace: Trace):
        _log(out, trace)

    return to_out



from treepath.path.exceptions.treepath_exception import TreepathException


class StopTraversing(TreepathException, StopIteration):
    """StopTraversing is raised when the data structure has no more element to traverse."""

    def _resolve_msg(self):
        path = repr(self.vertex)
        return f"StopTraversing(Traversing has completed on path: {path})"



from typing import Union

from treepath.path.traverser.match_traverser import MatchTraverser


class ValueTraverser(MatchTraverser):
    __slots__ = ()

    def __next__(self) -> Union[dict, list, str, int, float, bool, None]:
        return super().__next__().data



from abc import abstractmethod, ABC, abstractproperty
from typing import List, Any, Callable


class TraverserMatch(ABC):
    __slots__ = 'real_parent', \
                'real_data_name', \
                'data', \
                'vertex', \
                'vertex_index', \
                '_path_as_list', \
                '_path', \
                'remembered_catch_state', \
                'remembered_on_catch_match', \
                'remembered_on_catch_action'

    def __init__(self,
                 real_parent,
                 data_name,
                 data,
                 vertex,
                 vertex_index,
                 remembered_on_catch_match,
                 remembered_on_catch_action
                 ):
        self.real_parent = real_parent
        self.real_data_name = data_name
        self.data = data
        self.vertex = vertex
        self.vertex_index = vertex_index
        self._path_as_list = self
        self._path = self
        self.remembered_catch_state = None
        self.remembered_on_catch_match = remembered_on_catch_match
        self.remembered_on_catch_action = remembered_on_catch_action

    @property
    def path_as_list(self) -> List[Any]:
        path_as_list = self._path_as_list
        if path_as_list != self:
            return path_as_list

        path_as_list = []

        def collect(vertex):
            path_as_list.append(vertex)

        self.traverse(collect)

        self._path_as_list = path_as_list
        return path_as_list

    @property
    def path(self) -> str:
        path = self._path
        if path != self:
            return path

        path_as_list = self.path_as_list
        path = ''.join(match.path_segment for match in path_as_list)
        self._path = path
        return path

    @property
    @abstractmethod
    def path_segment(self) -> str:
        raise NotImplementedError

    @property
    def data_name(self) -> str:
        return self.real_data_name

    @property
    def parent(self):
        return self.real_parent

    @property
    def remembered_parent(self):
        return self.parent

    def traverse(self, visit: Callable[[Any], None]):
        self.real_parent.traverse(visit)
        visit(self)

    def __repr__(self):
        return f"{self.path}={self.data}"

    def __str__(self):
        real_parent = id(self.real_parent)
        data_name = self.real_data_name
        data = type(self.data).__name__
        vertex = type(self.vertex).__name__
        vertex_index = self.vertex_index
        remembered_catch_state = type(self.remembered_catch_state).__name__
        remembered_on_catch_match = id(self.remembered_on_catch_match)
        remembered_on_catch_action = type(self.remembered_on_catch_action).__name__
        path = self.path

        return f"self={id(self)}" \
               f" self={type(self).__name__}" \
               f" parent={real_parent}" \
               f" real_data_name={data_name}" \
               f" data={data}" \
               f" vertex={vertex}" \
               f" vertex_index={vertex_index}" \
               f" remembered_catch_state={remembered_catch_state}" \
               f" remembered_on_catch_match={remembered_on_catch_match}" \
               f" remembered_on_catch_action={remembered_on_catch_action}" \
               f" path={path}"



import typing

from treepath.path.traverser.traverser_match import TraverserMatch


class RootMatch(TraverserMatch):
    __slots__ = ()

    @property
    def path_as_list(self) -> list:
        return [self]

    @property
    def path(self) -> str:
        return self.real_data_name

    @property
    def path_segment(self) -> str:
        return self.real_data_name

    def traverse(self, visit: typing.Callable[[TraverserMatch], None]):
        visit(self)



from treepath.path.builder.path_builder import PathBuilder
from treepath.path.vertex.root_vertex import RootVertex


class DashPathBuilder(PathBuilder):
    __slots__ = ()

    def create_path_builder(self, *args, **kwargs):
        return DashPathBuilder(*args, **kwargs)

    def tranform_attribute_name(self, name):
        return name.replace('_', '-')


class DashPathRoot(DashPathBuilder):
    __slots__ = ()

    def __init__(self):
        vertex = RootVertex("$")
        super().__init__(vertex)



from treepath.path.exceptions.traversing_error import TraversingError


class InfiniteLoopDetected(TraversingError):
    """InfiniteLoopDetected is raised when the next is detect to be in a infinite loop."""

    def _resolve_msg(self):
        path = repr(self.vertex)
        return f"InfiniteLoopDetected(Traversing seems to go on for ever on path: {path})"



from treepath.path.traverser.match import Match
from treepath.path.traverser.trace import Trace


class PredicateMatch(Match):
    __slots__ = '_trace'

    def __init__(self, trace, *args):
        super().__init__(*args)
        self._trace = trace

    @property
    def trace(self):
        trace_function = self._trace
        if trace_function:
            def predicate_trace(trace: Trace):
                trace.predicate_match = self
                trace_function(trace)

            return predicate_trace
        else:
            return None
